from functools import partial

import numpy as np
import pyproj
from numpy import ma as ma
from shapely.ops import transform


def crossings(profile_x, profile_z, constant_z):
    """
    Finds the x coordinate of a z elevation for a beach profile. Much faster than using shapely to calculate
    intersections since we are only interested in intersections of a constant value. Will return multiple
    intersections if found. Used in calculating beach slope.
    Adapted from https://stackoverflow.com/a/34745789
    :param profile_x: List of x coordinates for the beach profile section
    :param profile_z: List of z coordinates for the beach profile section
    :param constant_z: Float of the elevation to find corresponding x coordinates
    :return: List of x coordinates which correspond to the constant_z
    """

    # Remove nans to suppress warning messages
    valid = ~ma.masked_invalid(profile_z).mask
    profile_z = np.array(profile_z)[valid]
    profile_x = np.array(profile_x)[valid]

    # Return empty list if mask removes all values
    if profile_x.size == 0:
        return []

    # Normalize the 'signal' to zero.
    # Use np.subtract rather than a list comprehension for performance reasons
    z = np.subtract(profile_z, constant_z)

    # Find all indices right before any crossing.
    # TODO Sometimes this can give a runtime warning https://stackoverflow.com/a/36489085
    indicies = np.where(z[:-1] * z[1:] < 0)[0]

    # Use linear interpolation to find intersample crossings.
    x_crossings = [
        profile_x[i] - (profile_x[i] - profile_x[i + 1]) / (z[i] - z[i + 1]) * (z[i])
        for i in indicies
    ]

    # Also need to check the end points as the above will not include them if they are close
    x_crossings += [
        profile_x[i]
        for i in [0, len(profile_z) - 1]
        if np.isclose(constant_z, profile_z[i])
    ]

    return sorted(x_crossings)


# TODO Think of a better way to do this than having to manually specify the coordinate systems
def convert_coord_systems(
    g1, in_coord_system="EPSG:4326", out_coord_system="EPSG:28356"
):
    """
    Converts coordinates from one coordinates system to another. Needed because shapefiles are usually defined in
    lat/lon but should be converted to GDA to calculated distances.
    https://gis.stackexchange.com/a/127432
    :param in_coord_system: Default is lat/lon WGS84
    :param out_coord_system: Default is GDA56 for NSW coastline
    :return:
    """
    project = partial(
        pyproj.transform,
        pyproj.Proj(init=in_coord_system),  # source coordinate system
        pyproj.Proj(init=out_coord_system),
    )  # destination coordinate system

    g2 = transform(project, g1)  # apply projection
    return g2


def get_i_or_default(l, i, default=None):
    try:
        return l[i]
    except IndexError:
        return default
